import { useState } from 'react'
import { Link } from 'react-router-dom'
import { Tooltip, Button, Divider } from '@mui/material'
import {
  faCalendar,
  faFile,
  faFileCircleExclamation,
  faPen,
  faTrash
} from '@fortawesome/free-solid-svg-icons'
import { FontAwesomeIcon } from '@fortawesome/react-fontawesome'
import { SigitDraft, UserRole } from '../../types'
import { appPrivateRoutes } from '../../routes'
import {
  formatTimestamp,
  getSigitDraft,
  npubToHex,
  SigitStatus,
  SignStatus
} from '../../utils'
import { DisplaySigner } from '../DisplaySigner'
import { UserAvatarGroup } from '../UserAvatarGroup'
import { getExtensionIconLabel } from '../getExtensionIconLabel'
import { useAppSelector, useDidMount } from '../../hooks'
import styles from './style.module.scss'

interface LocalDraftSigitProps {
  handleDraftDelete: () => void
}
export const LocalDraftSigit = ({
  handleDraftDelete
}: LocalDraftSigitProps) => {
  const [draft, setDraft] = useState<SigitDraft>()
  useDidMount(async () => {
    // Check if draft exists and add link to direct
    const draft = await getSigitDraft()
    if (draft) {
      setDraft(draft)
    }
  })
  const submittedBy = useAppSelector((state) => state.auth.usersPubkey)

  if (!draft) return null

  const extensions = draft.files.map((f) => f.extension)
  const isSame = extensions.every((e) => extensions[0] === e)

  return (
    <div className={styles.itemWrapper}>
      <Link className={styles.insetLink} to={appPrivateRoutes.create}></Link>
      <p className={`line-clamp-2 ${styles.title}`}>{draft.title}</p>
      <div className={styles.users}>
        {submittedBy && (
          <DisplaySigner status={SignStatus.Pending} pubkey={submittedBy} />
        )}
        {submittedBy && draft.users.length ? (
          <Divider orientation="vertical" flexItem />
        ) : null}
        <UserAvatarGroup max={7}>
          {draft.users.map((user) => {
            const pubkey = npubToHex(user.pubkey)!
            return (
              <DisplaySigner
                key={pubkey}
                status={
                  user.role === UserRole.signer
                    ? SignStatus.Pending
                    : SignStatus.Viewer
                }
                pubkey={pubkey}
              />
            )
          })}
        </UserAvatarGroup>
      </div>
      <div className={`${styles.details} ${styles.iconLabel}`}>
        <FontAwesomeIcon icon={faCalendar} />
        {formatTimestamp(draft.lastUpdated)}
      </div>
      <div className={`${styles.details} ${styles.status}`}>
        <span className={styles.iconLabel}>
          <FontAwesomeIcon icon={faPen} /> {SigitStatus.LocalDraft}
        </span>
        {extensions.length > 0 ? (
          <span className={styles.iconLabel}>
            {!isSame ? (
              <>
                <FontAwesomeIcon icon={faFile} /> Multiple File Types
              </>
            ) : (
              getExtensionIconLabel(extensions[0])
            )}
          </span>
        ) : (
          <>
            <FontAwesomeIcon icon={faFileCircleExclamation} /> &mdash;
          </>
        )}
      </div>
      <div className={styles.itemActions}>
        <Tooltip title="Delete" arrow placement="top" disableInteractive>
          <Button
            onClick={handleDraftDelete}
            sx={{
              color: 'var(--primary-main)',
              minWidth: '34px',
              padding: '10px'
            }}
            variant={'text'}
          >
            <FontAwesomeIcon icon={faTrash} />
          </Button>
        </Tooltip>
      </div>
    </div>
  )
}
