import React, { useMemo } from 'react'
import {
  getLocalStorageItem,
  mergeWithInitialValue,
  removeLocalStorageItem,
  setLocalStorageItem
} from '../utils'

/**
 * Subscribe to the Browser's storage event. Get the new value if any of the tabs changes it.
 * @param callback - function to be called when the storage event is triggered
 * @returns clean up function
 */
const useLocalStorageSubscribe = (callback: () => void) => {
  window.addEventListener('storage', callback)
  return () => window.removeEventListener('storage', callback)
}

export function useLocalStorage<T>(
  key: string,
  initialValue: T
): [T, React.Dispatch<React.SetStateAction<T>>] {
  const getSnapshot = () => {
    // Get the stored value
    const storedValue = getLocalStorageItem(key, initialValue)

    // Parse the value
    const parsedStoredValue = JSON.parse(storedValue)

    // Merge the default and the stored in case some of the required fields are missing
    return JSON.stringify(
      mergeWithInitialValue(parsedStoredValue, initialValue)
    )
  }

  // https://react.dev/reference/react/useSyncExternalStore
  // Returns the snapshot of the data and subscribes to the storage event
  const data = React.useSyncExternalStore(useLocalStorageSubscribe, getSnapshot)

  // Takes the value or a function that returns the value and updates the local storage
  const setState: React.Dispatch<React.SetStateAction<T>> = React.useCallback(
    (v: React.SetStateAction<T>) => {
      try {
        const nextState =
          typeof v === 'function'
            ? (v as (prevState: T) => T)(JSON.parse(data))
            : v

        if (nextState === undefined || nextState === null) {
          removeLocalStorageItem(key)
        } else {
          setLocalStorageItem(key, JSON.stringify(nextState))
        }
      } catch (e) {
        console.warn(e)
      }
    },
    [data, key]
  )

  React.useEffect(() => {
    // Set local storage only when it's empty
    const data = window.localStorage.getItem(key)
    if (data === null) {
      setLocalStorageItem(key, JSON.stringify(initialValue))
    }
  }, [key, initialValue])

  const memoized = useMemo(() => JSON.parse(data) as T, [data])

  return [memoized, setState]
}
