import axios from 'axios'
import JSZip from 'jszip'
import _ from 'lodash'
import { Event, verifyEvent } from 'nostr-tools'
import { useEffect, useMemo, useState } from 'react'
import { useAppSelector } from '../../hooks'
import { useLocation, useNavigate, useParams } from 'react-router-dom'
import { toast } from 'react-toastify'
import { LoadingSpinner } from '../../components/LoadingSpinner'
import { NostrController } from '../../controllers'
import { appPublicRoutes } from '../../routes'
import { CreateSignatureEventContent, Meta, SignedEvent } from '../../types'
import {
  decryptArrayBuffer,
  extractMarksFromSignedMeta,
  extractZipUrlAndEncryptionKey,
  filterMarksByPubkey,
  findOtherUserMarks,
  getCurrentUserFiles,
  getCurrentUserMarks,
  getHash,
  hexToNpub,
  loadZip,
  npubToHex,
  parseJson,
  encryptAndUploadMarks,
  readContentOfZipEntry,
  signEventForMetaFile,
  unixNow,
  updateMarks,
  uploadMetaToFileStorage,
  parseNostrEvent
} from '../../utils'
import { CurrentUserMark, Mark } from '../../types/mark.ts'
import PdfMarking from '../../components/PDFView/PdfMarking.tsx'
import { convertToSigitFile, SigitFile } from '../../utils/file.ts'
import { generateTimestamp } from '../../utils/opentimestamps.ts'
import { MARK_TYPE_CONFIG } from '../../components/MarkTypeStrategy/MarkStrategy.tsx'
import { useNDK } from '../../hooks/useNDK.ts'
import { SendDMError } from '../../types/errors/SendDMError.ts'

export const SignPage = () => {
  const navigate = useNavigate()
  const location = useLocation()
  const params = useParams()
  const { updateUsersAppData, sendNotification, sendPrivateDirectMessage } =
    useNDK()

  const usersAppData = useAppSelector((state) => state.userAppData)

  /**
   * In the online mode, Sigit ID can be obtained either from the router state
   * using location or from UsersAppData
   */
  const metaInNavState = useMemo(() => {
    if (usersAppData) {
      const sigitCreateId = params.id

      if (sigitCreateId) {
        const sigit = usersAppData.sigits[sigitCreateId]

        if (sigit) {
          return sigit
        }
      }
    }

    return location?.state?.meta || undefined
  }, [location, usersAppData, params.id])

  /**
   * Received from `location.state`
   *
   * arrayBuffer (decryptedArrayBuffer) will be received in navigation from create page in offline mode
   */
  const { arrayBuffer: decryptedArrayBuffer, uploadedZip } = location.state || {
    decryptedArrayBuffer: undefined
  }

  const [files, setFiles] = useState<{ [filename: string]: SigitFile }>({})

  const [isLoading, setIsLoading] = useState(true)
  const [loadingSpinnerDesc, setLoadingSpinnerDesc] = useState('')

  const [meta, setMeta] = useState<Meta | null>(null)

  const [submittedBy, setSubmittedBy] = useState<string>()

  const [signers, setSigners] = useState<`npub1${string}`[]>([])
  const [viewers, setViewers] = useState<`npub1${string}`[]>([])
  const [marks, setMarks] = useState<Mark[]>([])
  const [creatorFileHashes, setCreatorFileHashes] = useState<{
    [key: string]: string
  }>({})
  const [currentFileHashes, setCurrentFileHashes] = useState<{
    [key: string]: string | null
  }>({})

  const usersPubkey = useAppSelector((state) => state.auth.usersPubkey)

  const nostrController = NostrController.getInstance()
  const [currentUserMarks, setCurrentUserMarks] = useState<CurrentUserMark[]>(
    []
  )
  const [otherUserMarks, setOtherUserMarks] = useState<Mark[]>([])

  useEffect(() => {
    const handleUpdatedMeta = async (meta: Meta) => {
      const createSignatureEvent = await parseJson<Event>(
        meta.createSignature
      ).catch((err) => {
        console.log('err in parsing the createSignature event:>> ', err)
        toast.error(
          err.message || 'error occurred in parsing the create signature event'
        )
        setIsLoading(false)
        return null
      })

      if (!createSignatureEvent) return

      const isValidCreateSignature = verifyEvent(createSignatureEvent)

      if (!isValidCreateSignature) {
        toast.error('Create signature is invalid')
        setIsLoading(false)
        return
      }

      const createSignatureContent =
        await parseJson<CreateSignatureEventContent>(
          createSignatureEvent.content
        ).catch((err) => {
          console.log(
            `err in parsing the createSignature event's content :>> `,
            err
          )
          toast.error(
            err.message ||
              `error occurred in parsing the create signature event's content`
          )
          setIsLoading(false)
          return null
        })

      if (!createSignatureContent) return

      setSigners(createSignatureContent.signers)
      setViewers(createSignatureContent.viewers)
      setCreatorFileHashes(createSignatureContent.fileHashes)
      setSubmittedBy(createSignatureEvent.pubkey)
      setMarks(createSignatureContent.markConfig)

      if (usersPubkey) {
        const metaMarks = filterMarksByPubkey(
          createSignatureContent.markConfig,
          usersPubkey!
        )
        const signedMarks = extractMarksFromSignedMeta(meta)
        const currentUserMarks = getCurrentUserMarks(metaMarks, signedMarks)
        const otherUserMarks = findOtherUserMarks(signedMarks, usersPubkey!)

        if (meta.keys) {
          for (let i = 0; i < otherUserMarks.length; i++) {
            const m = otherUserMarks[i]
            const { sender, keys } = meta.keys
            const usersNpub = hexToNpub(usersPubkey)
            if (usersNpub in keys) {
              const encryptionKey = await nostrController
                .nip04Decrypt(sender, keys[usersNpub])
                .catch((err) => {
                  console.log(
                    'An error occurred in decrypting encryption key',
                    err
                  )
                  return null
                })

              try {
                const { fetchAndDecrypt } = MARK_TYPE_CONFIG[m.type] || {}
                if (
                  typeof fetchAndDecrypt === 'function' &&
                  m.value &&
                  encryptionKey
                ) {
                  otherUserMarks[i].value = await fetchAndDecrypt(
                    m.value,
                    encryptionKey
                  )
                }
              } catch (error) {
                console.error(`Error during mark fetchAndDecrypt phase`, error)
              }
            }
          }
        }

        setOtherUserMarks(otherUserMarks)
        setCurrentUserMarks(currentUserMarks)
      }
    }

    if (meta) {
      handleUpdatedMeta(meta)
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [meta, usersPubkey])

  useEffect(() => {
    if (metaInNavState) {
      const processSigit = async () => {
        setIsLoading(true)
        setLoadingSpinnerDesc('Extracting zipUrl and encryption key from meta')

        const res = await extractZipUrlAndEncryptionKey(metaInNavState)
        if (!res) {
          setIsLoading(false)
          return
        }

        const { zipUrl, encryptionKey } = res

        setLoadingSpinnerDesc('Fetching file from file server')
        axios
          .get(zipUrl, {
            responseType: 'arraybuffer'
          })
          .then((res) => {
            handleArrayBufferFromBlossom(res.data, encryptionKey)
            setMeta(metaInNavState)
          })
          .catch((err) => {
            console.error(`error occurred in getting file from ${zipUrl}`, err)
            toast.error(
              err.message || `error occurred in getting file from ${zipUrl}`
            )
          })
          .finally(() => {
            setIsLoading(false)
          })
      }

      processSigit()
    }

    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [])

  useEffect(() => {
    if (decryptedArrayBuffer || uploadedZip) {
      handleDecryptedArrayBuffer(decryptedArrayBuffer || uploadedZip).finally(
        () => setIsLoading(false)
      )
    } else {
      setIsLoading(false)
    }
  }, [decryptedArrayBuffer, uploadedZip])

  const handleArrayBufferFromBlossom = async (
    arrayBuffer: ArrayBuffer,
    encryptionKey: string
  ) => {
    // array buffer returned from blossom is encrypted.
    // So, first decrypt it
    const decrypted = await decryptArrayBuffer(
      arrayBuffer,
      encryptionKey
    ).catch((err) => {
      console.log('err in decryption:>> ', err)
      toast.error(err.message || 'An error occurred in decrypting file.')
      setIsLoading(false)
      return null
    })

    if (!decrypted) return

    const zip = await loadZip(decrypted)
    if (!zip) {
      setIsLoading(false)
      return
    }

    const files: { [filename: string]: SigitFile } = {}
    const fileHashes: { [key: string]: string | null } = {}
    const fileNames = Object.values(zip.files).map((entry) => entry.name)

    // generate hashes for all files in zipArchive
    // these hashes can be used to verify the originality of files
    for (const fileName of fileNames) {
      const arrayBuffer = await readContentOfZipEntry(
        zip,
        fileName,
        'arraybuffer'
      )

      if (arrayBuffer) {
        files[fileName] = await convertToSigitFile(arrayBuffer, fileName)
        const hash = await getHash(arrayBuffer)
        if (hash) {
          fileHashes[fileName] = hash
        }
      } else {
        fileHashes[fileName] = null
      }
    }

    setFiles(files)
    setCurrentFileHashes(fileHashes)
  }

  const setUpdatedMarks = (markToUpdate: Mark) => {
    const updatedMarks = updateMarks(marks, markToUpdate)
    setMarks(updatedMarks)
  }

  const handleDecryptedArrayBuffer = async (
    decryptedArrayBuffer: ArrayBuffer
  ) => {
    setLoadingSpinnerDesc('Parsing zip file')

    const zip = await loadZip(decryptedArrayBuffer)
    if (!zip) return

    const files: { [filename: string]: SigitFile } = {}
    const fileHashes: { [key: string]: string | null } = {}
    const fileNames = Object.values(zip.files)
      .filter((entry) => entry.name.startsWith('files/') && !entry.dir)
      .map((entry) => entry.name)

    for (const zipFilePath of fileNames) {
      const arrayBuffer = await readContentOfZipEntry(
        zip,
        zipFilePath,
        'arraybuffer'
      )

      const fileName = zipFilePath.replace(/^files\//, '')
      if (arrayBuffer) {
        files[fileName] = await convertToSigitFile(arrayBuffer, fileName)

        // generate hashes for all entries in files folder of zipArchive
        // these hashes can be used to verify the originality of files
        const hash = await getHash(arrayBuffer)
        if (hash) {
          fileHashes[fileName] = hash
        }
      } else {
        fileHashes[fileName] = null
      }
    }

    setFiles(files)
    setCurrentFileHashes(fileHashes)
    setLoadingSpinnerDesc('Parsing meta.json')

    const metaFileContent = await readContentOfZipEntry(
      zip,
      'meta.json',
      'string'
    )

    if (!metaFileContent) {
      setIsLoading(false)
      return
    }

    const parsedMetaJson = await parseJson<Meta>(metaFileContent).catch(
      (err) => {
        console.log('err in parsing the content of meta.json :>> ', err)
        toast.error(
          err.message || 'error occurred in parsing the content of meta.json'
        )
        setIsLoading(false)
        return null
      }
    )

    setMeta(parsedMetaJson)
  }

  const initializeSigning = async (type: 'online' | 'offline') => {
    if (Object.entries(files).length === 0 || !meta) return

    setIsLoading(true)
    setLoadingSpinnerDesc('Signing nostr event')

    const usersNpub = hexToNpub(usersPubkey!)
    const prevSig = getPrevSignersSig(usersNpub)
    if (!prevSig) {
      toast.error('Previous signature is invalid')
      return
    }

    const marks = getSignerMarksForMeta() || []

    let encryptionKey: string | undefined
    if (meta.keys) {
      const { sender, keys } = meta.keys
      encryptionKey = await nostrController
        .nip04Decrypt(sender, keys[usersNpub])
        .catch((err) => {
          // Log and display an error message if decryption fails
          console.log('An error occurred in decrypting encryption key', err)
          toast.error('An error occurred in decrypting encryption key')
          return undefined
        })
    }

    const processedMarks =
      type === 'online'
        ? await encryptAndUploadMarks(marks, encryptionKey)
        : marks

    const signedEvent = await signEventForMeta({
      prevSig,
      marks: processedMarks
    })

    if (!signedEvent) return

    const updatedMeta = updateMetaSignatures(meta, signedEvent)

    return {
      encryptionKey,
      updatedMeta,
      signedEvent
    }
  }

  const handleSign = async () => {
    const result = await initializeSigning('online')
    if (!result) {
      setIsLoading(false)
      return
    }

    const { encryptionKey, updatedMeta, signedEvent } = result

    setLoadingSpinnerDesc('Generating an open timestamp.')

    const timestamp = await generateTimestamp(signedEvent.id)
    if (timestamp) {
      updatedMeta.timestamps = [...(updatedMeta.timestamps || []), timestamp]
      updatedMeta.modifiedAt = unixNow()
    }

    await handleOnlineFlow(updatedMeta, encryptionKey)

    const createSignature = JSON.parse(updatedMeta.createSignature)
    navigate(`${appPublicRoutes.verify}/${createSignature.id}`)
  }

  const handleSignOffline = async () => {
    const result = await initializeSigning('offline')
    if (!result) {
      setIsLoading(false)
      return
    }

    const { updatedMeta } = result

    const zip = new JSZip()
    for (const [filename, value] of Object.entries(files)) {
      zip.file(`files/${filename}`, await value.arrayBuffer())
    }
    const stringifiedMeta = JSON.stringify(updatedMeta, null, 2)
    zip.file('meta.json', stringifiedMeta)

    // Handle errors during zip file generation
    const handleZipError = (err: unknown) => {
      console.log('Error in zip:>> ', err)
      setIsLoading(false)
      if (err instanceof Error) {
        toast.error(err.message || 'Error occurred in generating zip file')
      }
      return null
    }

    setLoadingSpinnerDesc('Generating zip file')

    const arrayBuffer = await zip
      .generateAsync({
        type: 'arraybuffer',
        compression: 'DEFLATE',
        compressionOptions: { level: 6 }
      })
      .catch(handleZipError)

    if (!arrayBuffer) {
      setIsLoading(false)
      return
    }

    // Create a File object with the Blob data
    const blob = new Blob([arrayBuffer])
    const file = new File([blob], `request-${unixNow()}.sigit.zip`, {
      type: 'application/zip'
    })

    setIsLoading(false)

    navigate(`${appPublicRoutes.verify}`, { state: { uploadedZip: file } })
  }

  // Sign the event for the meta file
  const signEventForMeta = async (signerContent: {
    prevSig: string
    marks: Mark[]
  }) => {
    return await signEventForMetaFile(
      JSON.stringify(signerContent),
      nostrController,
      setIsLoading
    )
  }

  const getSignerMarksForMeta = (): Mark[] | undefined => {
    if (currentUserMarks.length === 0) return
    return currentUserMarks.map(({ mark }: CurrentUserMark) => mark)
  }

  // Update the meta signatures
  const updateMetaSignatures = (meta: Meta, signedEvent: SignedEvent): Meta => {
    const metaCopy = _.cloneDeep(meta)
    metaCopy.docSignatures = {
      ...metaCopy.docSignatures,
      [hexToNpub(signedEvent.pubkey)]: JSON.stringify(signedEvent, null, 2)
    }
    metaCopy.modifiedAt = unixNow()
    return metaCopy
  }

  // Check if the current user is the last signer
  const checkIsLastSigner = (signers: string[]): boolean => {
    const usersNpub = hexToNpub(usersPubkey!)
    const lastSignerIndex = signers.length - 1
    const signerIndex = signers.indexOf(usersNpub)
    return signerIndex === lastSignerIndex
  }

  // Handle the online flow: update users app data and send notifications
  const handleOnlineFlow = async (
    meta: Meta,
    encryptionKey: string | undefined
  ) => {
    setLoadingSpinnerDesc('Updating users app data')
    const updatedEvent = await updateUsersAppData([meta])
    if (!updatedEvent) {
      setIsLoading(false)
      return
    }

    let metaUrl: string
    try {
      metaUrl = await uploadMetaToFileStorage(meta, encryptionKey)
    } catch (error) {
      if (error instanceof Error) {
        toast.error(error.message)
      }
      console.error(error)
      setIsLoading(false)
      return
    }

    const userSet = new Set<`npub1${string}`>()
    if (submittedBy && submittedBy !== usersPubkey) {
      userSet.add(hexToNpub(submittedBy))
    }

    const usersNpub = hexToNpub(usersPubkey!)
    const isLastSigner = checkIsLastSigner(signers)
    if (isLastSigner) {
      signers.forEach((signer) => {
        if (signer !== usersNpub) {
          userSet.add(signer)
        }
      })

      viewers.forEach((viewer) => {
        userSet.add(viewer)
      })
    } else {
      const currentSignerIndex = signers.indexOf(usersNpub)
      const prevSigners = signers.slice(0, currentSignerIndex)

      prevSigners.forEach((signer) => {
        userSet.add(signer)
      })

      const nextSigner = signers[currentSignerIndex + 1]
      userSet.add(nextSigner)
    }

    setLoadingSpinnerDesc('Sending notifications')
    const users = Array.from(userSet)
    const promises = users.map((user) =>
      sendNotification(npubToHex(user)!, { metaUrl, keys: meta.keys })
    )
    await Promise.all(promises)
      .then(() => {
        toast.success('Notifications sent successfully')
        setMeta(meta)
      })
      .catch(() => {
        toast.error('Failed to publish notifications')
      })

    // Send DMs
    setLoadingSpinnerDesc('Sending DMs')
    const createSignatureEvent = parseNostrEvent(meta.createSignature)
    const { id } = createSignatureEvent

    if (isLastSigner) {
      // Final sign sends to everyone (creator, signers, viewers - /verify)
      const areSent: boolean[] = Array(users.length).fill(false)
      for (let i = 0; i < users.length; i++) {
        try {
          areSent[i] = await sendPrivateDirectMessage(
            `Sigit completed, visit ${window.location.origin}/#/verify/${id}`,
            npubToHex(users[i])!
          )
        } catch (error) {
          if (error instanceof SendDMError) {
            toast.error(error.message)
          }
          console.error(error)
        }
      }

      if (areSent.some((r) => r)) {
        toast.success(
          `DMs sent ${areSent.filter((r) => r).length}/${users.length}`
        )
      }
    } else {
      // Notify the creator and
      // the next signer (/sign).
      try {
        await sendPrivateDirectMessage(
          `Sigit signed by ${usersNpub}, visit ${window.location.origin}/#/sign/${id}`,
          npubToHex(submittedBy!)!
        )
      } catch (error) {
        if (error instanceof SendDMError) {
          toast.error(error.message)
        }
        console.error(error)
      }

      // No need to notify creator twice, skipping
      const currentSignerIndex = signers.indexOf(usersNpub)
      const nextSigner = npubToHex(signers[currentSignerIndex + 1])
      if (nextSigner !== submittedBy) {
        try {
          await sendPrivateDirectMessage(
            `You're the next signer, visit ${window.location.origin}/#/sign/${id}`,
            nextSigner!
          )
        } catch (error) {
          if (error instanceof SendDMError) {
            toast.error(error.message)
          }
          console.error(error)
        }
      }
    }

    setIsLoading(false)
  }

  /**
   * This function accepts an npub of a signer and return the signature of its previous signer.
   * This prevSig will be used in the content of the provided signer's signedEvent
   */
  const getPrevSignersSig = (npub: string) => {
    if (!meta) return null

    // if user is first signer then use creator's signature
    if (signers[0] === npub) {
      try {
        const createSignatureEvent: Event = JSON.parse(meta.createSignature)
        return createSignatureEvent.sig
      } catch (error) {
        return null
      }
    }

    // find the index of signer
    const currentSignerIndex = signers.findIndex((signer) => signer === npub)
    // return null if could not found user in signer's list
    if (currentSignerIndex === -1) return null
    // find prev signer
    const prevSigner = signers[currentSignerIndex - 1]

    // get the signature of prev signer
    try {
      const prevSignersEvent: Event = JSON.parse(meta.docSignatures[prevSigner])
      return prevSignersEvent.sig
    } catch (error) {
      return null
    }
  }

  if (isLoading) {
    return <LoadingSpinner desc={loadingSpinnerDesc} />
  }

  return (
    <PdfMarking
      files={getCurrentUserFiles(files, currentFileHashes, creatorFileHashes)}
      currentUserMarks={currentUserMarks}
      setCurrentUserMarks={setCurrentUserMarks}
      setUpdatedMarks={setUpdatedMarks}
      handleSign={handleSign}
      handleSignOffline={handleSignOffline}
      otherUserMarks={otherUserMarks}
      meta={meta}
    />
  )
}
