import {
  DrawnField,
  SerializedSigitDraft,
  SigitDraft,
  SigitFileDraft
} from '../types'
import {
  getMediaType,
  extractFileExtension,
  toFile,
  getSigitFile
} from './file'
const DRAFT_KEY = 'sigitDraft'
let saveSigitDraftTimeout: number | null = null
const serializeSigitDraft = async (
  draft: SigitDraft
): Promise<SerializedSigitDraft> => {
  const serializedFiles = draft.files.map((file) => {
    return new Promise<SigitFileDraft>((resolve, reject) => {
      const reader = new FileReader()
      reader.onload = () => {
        const pages = file.pages
          ? file.pages.map((page) =>
              page.drawnFields.map(
                (field) =>
                  ({
                    left: field.left,
                    top: field.top,
                    width: field.width,
                    height: field.height,
                    type: field.type,
                    counterpart: field.counterpart
                  }) as DrawnField
              )
            )
          : []
        resolve({
          name: file.name,
          pages: pages,
          file: reader.result as string
        })
      }
      reader.onerror = (error) => reject(error)
      reader.readAsDataURL(file)
    })
  })

  const serializedFileDraft = await Promise.all(serializedFiles)
  return {
    title: draft.title,
    lastUpdated: draft.lastUpdated,
    users: [...draft.users],
    files: serializedFileDraft
  }
}

const deserializeSigitDraft = async (
  serializedDraft: SerializedSigitDraft
): Promise<SigitDraft> => {
  const files = await Promise.all(
    serializedDraft.files.map(async (draft) => {
      const response = await fetch(draft.file)
      const arrayBuffer = await response.arrayBuffer()
      const type = getMediaType(extractFileExtension(draft.name))
      const file = toFile(arrayBuffer, draft.name, type)
      const sigitFile = await getSigitFile(file)
      if (draft.pages) {
        for (let i = 0; i < draft.pages.length; i++) {
          const drawnFields = draft.pages[i]
          if (sigitFile.pages) sigitFile.pages[i].drawnFields = [...drawnFields]
        }
      }
      return sigitFile
    })
  )

  return {
    ...serializedDraft,
    files: files
  }
}

export const saveSigitDraft = (draft: SigitDraft): Promise<void> => {
  if (saveSigitDraftTimeout) {
    clearTimeout(saveSigitDraftTimeout)
  }

  return new Promise((resolve, reject) => {
    saveSigitDraftTimeout = window.setTimeout(() => {
      serializeSigitDraft(draft)
        .then((draftToSave) => {
          localStorage.setItem(DRAFT_KEY, JSON.stringify(draftToSave))
          resolve()
        })
        .catch((error) => {
          reject(error)
        })
    }, 1000)
  })
}

export const hasSigitDraft = () => {
  return DRAFT_KEY in localStorage
}

export const getSigitDraft = async () => {
  const sigitDraft = localStorage.getItem(DRAFT_KEY)
  if (!sigitDraft) return null

  try {
    const serializedDraft = JSON.parse(sigitDraft) as SerializedSigitDraft
    return await deserializeSigitDraft(serializedDraft)
  } catch {
    return null
  }
}

export const clearSigitDraft = () => {
  localStorage.removeItem(DRAFT_KEY)
}
