import JSZip from 'jszip'
import { toast } from 'react-toastify'
import { InputFileFormat, Meta, OutputByType, OutputType } from '../types'
import { NavigateOptions, To } from 'react-router-dom'
import { appPublicRoutes } from '../routes'
import { NostrController } from '../controllers'
import { decryptArrayBuffer } from './crypto'
import { hexToNpub, parseJson, SigitStatus, timeout } from '.'
import { SignerService } from '../services'

/**
 * Read the content of a file within a zip archive.
 * @param zip The JSZip object representing the zip archive.
 * @param filePath The path of the file within the zip archive.
 * @param outputType The type of output to return (e.g., 'string', 'arraybuffer', 'uint8array', etc.).
 * @returns A Promise resolving to the content of the file, or null if an error occurs.
 */
export const readContentOfZipEntry = async <T extends OutputType>(
  zip: JSZip,
  filePath: string,
  outputType: T
): Promise<OutputByType[T] | null> => {
  // Get the zip entry corresponding to the specified file path
  const zipEntry = zip.files[filePath]

  if (!zipEntry) {
    toast.error(`Couldn't find file in zip archive at ${filePath}`)
    return null
  }

  // Read and return the content of the zip entry asynchronously
  // or null if an error has occurred
  return await zipEntry.async(outputType).catch((err) => {
    // Handle any errors that occur during the read operation
    console.log(`Error reading content of ${filePath}:`, err)
    toast.error(
      err.message || `Error occurred in reading content of ${filePath}`
    )
    return null
  })
}

export const loadZip = async (data: InputFileFormat): Promise<JSZip | null> => {
  try {
    return await JSZip.loadAsync(data)
  } catch (err) {
    console.log('err in loading zip file :>> ', err)
    if (err instanceof Error) {
      toast.error(err.message || 'An error occurred in loading zip file.')
    }
    return null
  }
}

export const decrypt = async (file: File) => {
  const nostrController = NostrController.getInstance()

  const zip = await loadZip(file)
  if (!zip) return

  const keysFileContent = await readContentOfZipEntry(
    zip,
    'keys.json',
    'string'
  )

  if (!keysFileContent) return null

  const parsedKeysJson = await parseJson<{ sender: string; keys: string[] }>(
    keysFileContent
  ).catch((err) => {
    console.log(`Error parsing content of keys.json:`, err)
    toast.error(err.message || `Error parsing content of keys.json`)
    return null
  })

  if (!parsedKeysJson) return

  const encryptedArrayBuffer = await readContentOfZipEntry(
    zip,
    'compressed.sigit',
    'arraybuffer'
  )

  if (!encryptedArrayBuffer) return

  const { keys, sender } = parsedKeysJson

  for (const key of keys) {
    // decrypt the encryptionKey, with timeout (duration = 60 seconds)
    const encryptionKey = await Promise.race([
      nostrController.nip04Decrypt(sender, key),
      timeout(60000)
    ])
      .then((res) => {
        return res
      })
      .catch((err) => {
        console.log('err :>> ', err)
        return null
      })

    // Return if encryption failed
    if (!encryptionKey) continue

    const arrayBuffer = await decryptArrayBuffer(
      encryptedArrayBuffer,
      encryptionKey
    ).catch((err) => {
      console.log('err in decryption:>> ', err)
      return null
    })

    if (arrayBuffer) return arrayBuffer
  }

  return null
}

type NavigateArgs = { to: To; options?: NavigateOptions }
export const navigateFromZip = async (file: File, pubkey: `npub1${string}`) => {
  if (!file.name.endsWith('.sigit.zip')) {
    toast.error(`Not a SiGit zip file: ${file.name}`)
  }

  try {
    let zip = await JSZip.loadAsync(file)
    if (!zip) {
      return null
    }

    let arrayBuffer: ArrayBuffer | undefined
    if ('keys.json' in zip.files) {
      // Decrypt
      const decryptedArrayBuffer = await decrypt(file).catch((err) => {
        console.error(`error occurred in decryption`, err)
        toast.error(err.message || `error occurred in decryption`)
      })

      if (decryptedArrayBuffer) {
        // Replace the zip and continue processing
        zip = await JSZip.loadAsync(decryptedArrayBuffer)
        arrayBuffer = decryptedArrayBuffer
      }
    }

    if ('meta.json' in zip.files) {
      // Check where we need to navigate
      // Find Meta and process it for signer state
      const metaContent = await readContentOfZipEntry(
        zip,
        'meta.json',
        'string'
      )
      if (metaContent) {
        const meta = JSON.parse(metaContent) as Meta
        const signerService = new SignerService(meta)

        const to =
          signerService.getSignedStatus() === SigitStatus.Complete
            ? appPublicRoutes.verify
            : signerService.getNavigate(hexToNpub(pubkey))

        return {
          to,
          options: {
            state: { uploadedZip: arrayBuffer || file }
          }
        } as NavigateArgs
      }
    }

    return null
  } catch (err) {
    console.error('err in processing sigit zip file :>> ', err)
    if (err instanceof Error) {
      toast.error(err.message || 'An error occurred in loading zip file.')
    }
    return null
  }
}
